/**
 * Ext GWT 3.0.0-beta3 - Ext for GWT
 * Copyright(c) 2007-2011, Sencha, Inc.
 * licensing@sencha.com
 *
 * http://sencha.com/license
 */
package com.sencha.gxt.desktop.client;

import java.util.ArrayList;
import java.util.List;

import com.google.gwt.core.client.Scheduler;
import com.google.gwt.core.client.Scheduler.ScheduledCommand;
import com.google.gwt.resources.client.ImageResource;
import com.google.gwt.user.client.ui.HasWidgets;
import com.sencha.gxt.core.client.util.Margins;
import com.sencha.gxt.core.client.util.Padding;
import com.sencha.gxt.desktop.client.desktopapp.DesktopLayoutType;
import com.sencha.gxt.desktop.client.layout.CascadeDesktopLayout;
import com.sencha.gxt.desktop.client.layout.CenterDesktopLayout;
import com.sencha.gxt.desktop.client.layout.DesktopLayout;
import com.sencha.gxt.desktop.client.layout.TileDesktopLayout;
import com.sencha.gxt.widget.core.client.Window;
import com.sencha.gxt.widget.core.client.container.BoxLayoutContainer.BoxLayoutData;
import com.sencha.gxt.widget.core.client.container.VBoxLayoutContainer;
import com.sencha.gxt.widget.core.client.container.VBoxLayoutContainer.VBoxLayoutAlign;
import com.sencha.gxt.widget.core.client.container.VerticalLayoutContainer;
import com.sencha.gxt.widget.core.client.container.VerticalLayoutContainer.VerticalLayoutData;
import com.sencha.gxt.widget.core.client.container.Viewport;
import com.sencha.gxt.widget.core.client.event.ActivateEvent;
import com.sencha.gxt.widget.core.client.event.ActivateEvent.ActivateHandler;
import com.sencha.gxt.widget.core.client.event.DeactivateEvent;
import com.sencha.gxt.widget.core.client.event.DeactivateEvent.DeactivateHandler;
import com.sencha.gxt.widget.core.client.event.HideEvent;
import com.sencha.gxt.widget.core.client.event.HideEvent.HideHandler;
import com.sencha.gxt.widget.core.client.event.MinimizeEvent;
import com.sencha.gxt.widget.core.client.event.MinimizeEvent.MinimizeHandler;
import com.sencha.gxt.widget.core.client.event.ShowEvent;
import com.sencha.gxt.widget.core.client.event.ShowEvent.ShowHandler;

/**
 * A desktop represents a desktop like application which contains a task bar,
 * start menu, and shortcuts.
 * <p/>
 * Rather than adding content directly to the root panel, content should be
 * wrapped in windows. Windows can be opened via shortcuts and the start menu.
 */
public class Desktop {

  private class WindowHandler implements ActivateHandler<Window>, DeactivateHandler<Window>, MinimizeHandler,
      HideHandler, ShowHandler {

    @Override
    public void onActivate(ActivateEvent<Window> event) {
      markActive((Window) event.getSource());
    }

    @Override
    public void onDeactivate(DeactivateEvent<Window> event) {
      markInactive((Window) event.getSource());
    }

    @Override
    public void onHide(HideEvent event) {
      hideWindow((Window) event.getSource());
    }

    @Override
    public void onMinimize(MinimizeEvent event) {
      minimizeWindow((Window) event.getSource());
    }

    @Override
    public void onShow(ShowEvent event) {
      showWindow((Window) event.getSource());
    }
  }

  // TODO: Consider replacing getData/setData with strongly typed properties
  private static final String MINIMIZE_KEY = "minimize";
  private static final String TASKBUTTON_KEY = "taskButton";

  private VBoxLayoutContainer desktop;
  private TaskBar taskBar;
  private List<Shortcut> shortcuts;
  private List<Window> windows;
  private WindowHandler handler;
  private Window activeWindow;
  private VerticalLayoutContainer desktopContainer;
  private Viewport desktopViewport;
  private DesktopLayout desktopLayout;

  /**
   * Creates a new Desktop window.
   */
  public Desktop() {
  }

  public void activate(Window window) {
    if (!getWindows().contains(window)) {
      addWindow(window);
    }
    if (window != null && !window.isVisible()) {
      window.show();
    } else {
      window.toFront();
    }
  }

  /**
   * Adds a shortcut to the desktop.
   * 
   * @param shortcut the shortcut to add
   */
  public void addShortcut(Shortcut shortcut) {
    getShortcuts().add(shortcut);
    getDesktop().add(shortcut, new BoxLayoutData(new Margins(5)));
  }

  public void addStartMenuItem(StartMainMenuItem menuItem) {
    getTaskBar().getStartMenu().add(menuItem);
  }

  public void addToolMenuItem(StartToolMenuItem startToolMenuItem) {
    getTaskBar().getStartMenu().addTool(startToolMenuItem);
  }

  public void addToolSeparator() {
    getTaskBar().getStartMenu().addToolSeparator();
  }

  /**
   * Adds a window to the desktop.
   * 
   * @param window the window to add
   */
  public void addWindow(Window window) {
    if (windows.add(window)) {
      window.setContainer(getDesktopContainer().getElement());
      addWindowHandler(window, getHandler());
      // callback via onShow
    }
  }

  /**
   * Returns the container of the "desktop", which is the area that contains the
   * shortcuts (i.e. minus the task bar).
   * 
   * @return the desktop layout container
   */
  public VBoxLayoutContainer getDesktop() {
    if (desktop == null) {
      desktop = new VBoxLayoutContainer();
      // TODO: See review comments for setting wallpaper properly
      desktop.addStyleName("x-desktop");
      desktop.setPadding(new Padding(5));
      desktop.setVBoxLayoutAlign(VBoxLayoutAlign.STRETCHMAX);
    }
    return desktop;
  }

  public VerticalLayoutContainer getDesktopContainer() {
    if (desktopContainer == null) {
      desktopContainer = new VerticalLayoutContainer() {
        @Override
        public void onResize() {
          super.onResize();
          layoutDesktopCheck();
        }

        @Override
        protected void doLayout() {
          super.doLayout();
          layoutDesktopCheck();
        }
      };
      desktopContainer.add(getDesktop(), new VerticalLayoutData(-1, 1));
      desktopContainer.add(getTaskBar(), new VerticalLayoutData(1, -1));
    }
    return desktopContainer;
  }

  public Viewport getDesktopViewport() {
    if (desktopViewport == null) {
      desktopViewport = new Viewport();
      desktopViewport.add(getDesktopContainer());
    }
    return desktopViewport;
  }

  /**
   * Returns a list of the desktop's shortcuts.
   * 
   * @return the shortcuts
   */
  public List<Shortcut> getShortcuts() {
    if (shortcuts == null) {
      shortcuts = new ArrayList<Shortcut>();
    }
    return shortcuts;

  }

  /**
   * Returns the start menu.
   * 
   * @return the start menu
   */
  public StartMenu getStartMenu() {
    return taskBar.getStartMenu();
  }

  /**
   * Returns the desktop's task bar.
   * 
   * @return the task bar
   */
  public TaskBar getTaskBar() {
    if (taskBar == null) {
      taskBar = new TaskBar();
    }
    return taskBar;
  }

  /**
   * Returns a list of the desktop's windows.
   * 
   * @return the windows
   */
  public List<Window> getWindows() {
    if (windows == null) {
      windows = new ArrayList<Window>();
    }
    return windows;
  }

  public void go(HasWidgets container) {
    container.add(getDesktopViewport());
  }

  public void layoutDesktopCheck() {
    if (!isMaximized()) {
      getDesktopLayout().layoutDesktop(getDesktop().getElement(), getWindows(), getDesktop().getOffsetWidth(),
          getDesktop().getOffsetHeight());
    }
  }

  public void layoutDesktopCheckScheduled() {
    if (!isMaximized()) {
      scheduleDesktopLayout();
    }
  }

  /**
   * Minimizes the window.
   * 
   * @param window the window to minimize
   */
  public void minimizeWindow(Window window) {
    window.setData(MINIMIZE_KEY, true);
    window.hide();
  }

  /**
   * Removes a shortcut from the desktop.
   * 
   * @param shortcut the shortcut to remove
   */
  public void removeShortcut(Shortcut shortcut) {
    getShortcuts().remove(shortcut);
    getDesktop().remove(shortcut);
  }

  /**
   * Removes a window from the desktop.
   * 
   * @param window the window to remove
   */
  public void removeWindow(Window window) {
    if (windows.remove(window)) {
      removeWindowHandler(window, getHandler());
      if (activeWindow == window) {
        activeWindow = null;
      }
      taskBar.removeTaskButton((TaskButton) window.getData(TASKBUTTON_KEY));
    }
  }

  /**
   * Removes a previously added listener.
   * 
   * @param window the window
   * @param handler the listener to remove
   */
  public void removeWindowHandler(Window window, WindowHandler handler) {
    window.removeHandler(ActivateEvent.getType(), handler);
    window.removeHandler(DeactivateEvent.getType(), handler);
    window.removeHandler(MinimizeEvent.getType(), handler);
    window.removeHandler(HideEvent.getType(), handler);
    window.removeHandler(ShowEvent.getType(), handler);
  }

  public void setDesktopLayoutType(DesktopLayoutType desktopLayoutType) {
    switch (desktopLayoutType) {
      case TILE:
        desktopLayout = new TileDesktopLayout();
        break;
      case CASCADE:
        desktopLayout = new CascadeDesktopLayout();
        break;
      case CENTER:
        desktopLayout = new CenterDesktopLayout();
        break;
    }
  }

  public void setStartMenuHeading(String heading) {
    taskBar.getStartMenu().setHeading(heading);
  }

  public void setStartMenuIcon(ImageResource icon) {
    taskBar.getStartMenu().setIcon(icon);
  }

  /**
   * Adds a handler to receive window events.
   * 
   * @param window the window
   * @param handler the handler
   */
  private void addWindowHandler(Window window, WindowHandler handler) {
    window.addActivateHandler(handler);
    window.addDeactivateHandler(handler);
    window.addMinimizeHandler(handler);
    window.addHideHandler(handler);
    window.addShowHandler(handler);
  }

  private DesktopLayout getDesktopLayout() {
    if (desktopLayout == null) {
      desktopLayout = new TileDesktopLayout();
    }
    return desktopLayout;
  }

  private WindowHandler getHandler() {
    if (handler == null) {
      handler = new WindowHandler();
    }
    return handler;
  }

  private void hideWindow(Window window) {
    if (window.getData(MINIMIZE_KEY) != null) {
      markInactive(window);
      return;
    }
    if (activeWindow == window) {
      activeWindow = null;
    }
    taskBar.removeTaskButton((TaskButton) window.getData(TASKBUTTON_KEY));
    windows.remove(window);
    layoutDesktopCheck();
  }

  private boolean isMaximized() {
    for (Window window : getWindows()) {
      if (window.isVisible() && window.isMaximized()) {
        return true;
      }
    }
    return false;
  }

  private void markActive(Window window) {
    if (activeWindow != null && activeWindow != window) {
      markInactive(activeWindow);
    }
    taskBar.setActiveButton((TaskButton) window.getData(TASKBUTTON_KEY));
    activeWindow = window;
    TaskButton btn = window.getData(TASKBUTTON_KEY);
    btn.setValue(true);
    window.setData(MINIMIZE_KEY, null);
  }

  private void markInactive(Window window) {
    if (window == activeWindow) {
      activeWindow = null;
      TaskButton btn = window.getData(TASKBUTTON_KEY);
      btn.setValue(false);
    }
  }

  private void scheduleDesktopLayout() {
    Scheduler.get().scheduleDeferred(new ScheduledCommand() {
      @Override
      public void execute() {
        getDesktopLayout().layoutDesktop(getDesktop().getElement(), getWindows(), getDesktop().getOffsetWidth(),
            getDesktop().getOffsetHeight());
      }
    });
  }

  private void showWindow(Window window) {
    TaskButton btn = window.getData(TASKBUTTON_KEY);
    window.setData(MINIMIZE_KEY, null);
    if (btn != null && taskBar.getButtons().contains(btn)) {
      return;
    }
    btn = taskBar.addTaskButton(window);
    window.setData(TASKBUTTON_KEY, btn);
    layoutDesktopCheck();
  }

}
