/**
 * Ext GWT 3.0.0-beta3 - Ext for GWT
 * Copyright(c) 2007-2011, Sencha, Inc.
 * licensing@sencha.com
 *
 * http://sencha.com/license
 */
package com.sencha.gxt.desktop.client.desktopapp;

import java.util.Date;

import com.google.gwt.core.client.GWT;
import com.google.gwt.i18n.client.DateTimeFormat;
import com.google.gwt.i18n.client.DateTimeFormat.PredefinedFormat;
import com.google.gwt.storage.client.Storage;
import com.google.gwt.storage.client.StorageEvent;
import com.google.gwt.storage.client.StorageEvent.Handler;
import com.google.web.bindery.autobean.shared.AutoBean;
import com.google.web.bindery.autobean.shared.AutoBeanCodex;
import com.google.web.bindery.autobean.shared.AutoBeanUtils;
import com.google.web.bindery.event.shared.Event;
import com.sencha.gxt.data.shared.event.StoreAddEvent;
import com.sencha.gxt.data.shared.event.StoreAddEvent.StoreAddHandler;
import com.sencha.gxt.data.shared.event.StoreRemoveEvent;
import com.sencha.gxt.data.shared.event.StoreRemoveEvent.StoreRemoveHandler;
import com.sencha.gxt.data.shared.event.StoreUpdateEvent;
import com.sencha.gxt.data.shared.event.StoreUpdateEvent.StoreUpdateHandler;
import com.sencha.gxt.desktop.client.desktopapp.event.AddFileModelEvent;
import com.sencha.gxt.desktop.client.desktopapp.event.RemoveFileModelEvent;
import com.sencha.gxt.desktop.client.desktopapp.event.UpdateFileModelEvent;
import com.sencha.gxt.desktop.client.desktopapp.event.UpdateSettingsEvent;
import com.sencha.gxt.desktop.client.desktopapp.event.UpdateSettingsHandler;
import com.sencha.gxt.desktop.client.desktopapp.persistence.BackingStore;
import com.sencha.gxt.desktop.client.desktopapp.persistence.FileModel;
import com.sencha.gxt.desktop.client.desktopapp.persistence.FileModelFactory;
import com.sencha.gxt.desktop.client.desktopapp.persistence.FileModelProperties;
import com.sencha.gxt.desktop.client.desktopapp.persistence.FileSystem;
import com.sencha.gxt.desktop.client.desktopapp.persistence.HasMoveSupport;
import com.sencha.gxt.desktop.client.desktopapp.persistence.StorageBackingStore;

public class DesktopAppPresenter implements Presenter {

  private static final String STORAGE_KEY_PREFIX = "com.sencha.gxt.desktop.client.desktopapp";
  private static final String CURRENT_USER_KEY = "currentUser";
  private static final String SETTINGS_KEY = "settings";

  private DesktopBus desktopBus;

  private FileSystem fileSystem;
  private FileModelFactory fileModelFactory;
  private FileModelProperties fileModelProperties;
  private Storage storage;
  private BackingStore backingStore;
  private Settings settings;
  private SettingsAutoBeanFactory settingsAutoBeanFactory;
  private String currentUser;
  private Boolean isLocalStorageSupported;

  public DesktopAppPresenter(DesktopBus desktopBus) {
    this.desktopBus = desktopBus;
    desktopBus.addUpdateSettingsHandler(new UpdateSettingsHandler() {
      @Override
      public void onUpdateSettings(UpdateSettingsEvent event) {
        setSettings(getCurrentUser(), event.getSettings());
      }
    });
  }

  @Override
  public void clearLocalStorage() {
    System.err.println("DesktopAppPresenter.clearLocalStorage: clearing local storage");
    getBackingStore().clear();
  }

  @Override
  public void createNewUser(String userName) {
    if (isExistingUser(userName)) {
      throw new IllegalStateException("user " + userName + " already exists");
    }
    AutoBean<Settings> settingsAutoBean = getSettingsAutoBeanFactory().settings();
    Settings settings = settingsAutoBean.as();
    initializeSettings(settings);
    setSettings(userName, settings);
  }

  @Override
  public String getCurrentUser() {
    if (currentUser == null) {
      currentUser = getBackingStore().getItem(CURRENT_USER_KEY);
    }
    return currentUser;
  }

  @Override
  public FileSystem getFileSystem() {
    if (fileSystem == null) {
      fileSystem = new FileSystem(getFileModelProperties(), getBackingStore(), getFileModelFactory(), FileModel.class);
      fileSystem.addStoreAddHandler(new StoreAddHandler<FileModel>() {
        @Override
        public void onAdd(StoreAddEvent<FileModel> event) {
          for (FileModel fileModel : event.getItems()) {
            getDesktopBus().fireAddFileModelEvent(new AddFileModelEvent(fileModel, isMoveEvent(event)));
          }
        }
      });
      fileSystem.addStoreUpdateHandler(new StoreUpdateHandler<FileModel>() {
        @Override
        public void onUpdate(StoreUpdateEvent<FileModel> event) {
          for (FileModel fileModel : event.getItems()) {
            getDesktopBus().fireUpdateFileModelEvent(new UpdateFileModelEvent(fileModel));
          }
        }
      });
      fileSystem.addStoreRemoveHandler(new StoreRemoveHandler<FileModel>() {
        @Override
        public void onRemove(StoreRemoveEvent<FileModel> event) {
          getDesktopBus().fireRemoveFileModelEvent(new RemoveFileModelEvent(event.getItem(), isMoveEvent(event)));
        }
      });
    }
    return fileSystem;
  }

  @Override
  public Settings getSettings(String userName) {
    if (settings == null) {
      String payload = getBackingStore().getItem(getUserKey(userName, SETTINGS_KEY));
      settings = AutoBeanCodex.decode(getSettingsAutoBeanFactory(), Settings.class, payload).as();
    }
    return settings;
  }

  @Override
  public boolean isExistingUser(String userName) {
    return getBackingStore().getItem(getUserKey(userName, SETTINGS_KEY)) != null;
  }

  @Override
  public boolean isLocalStorageSupported() {
    if (!isLocalStorageSupportChecked()) {
      isLocalStorageSupported = Storage.isLocalStorageSupported();
    }
    return isLocalStorageSupported;
  }

  @Override
  public boolean isLoggedIn() {
    return getCurrentUser() != null;
  }

  @Override
  public void login(String currentUser) {
    if (!isExistingUser(currentUser)) {
      throw new IllegalArgumentException("user " + currentUser + " does not exist");
    }
    this.currentUser = currentUser;
    getBackingStore().setItem(CURRENT_USER_KEY, currentUser);
  }

  @Override
  public void logout() {
    currentUser = null;
    getBackingStore().removeItem(CURRENT_USER_KEY);
    com.google.gwt.user.client.Window.Location.reload();
  }

  @Override
  public void setSettings(String userName, Settings settings) {
    this.settings = settings;
    String payload = AutoBeanCodex.encode(AutoBeanUtils.getAutoBean(settings)).getPayload();
    getBackingStore().setItem(getUserKey(userName, SETTINGS_KEY), payload);
  }

  private BackingStore getBackingStore() {
    if (backingStore == null) {
      if (!isLocalStorageSupportChecked()) {
        throw new IllegalStateException("Application did not check local storage support");
      }
      if (!isLocalStorageSupported()) {
        backingStore = new MemoryBackingStore();
      } else {
        backingStore = new StorageBackingStore(getStorage(), STORAGE_KEY_PREFIX);
      }
    }
    return backingStore;
  }

  private DesktopBus getDesktopBus() {
    return desktopBus;
  }

  private FileModelFactory getFileModelFactory() {
    if (fileModelFactory == null) {
      fileModelFactory = GWT.create(FileModelFactory.class);
    }
    return fileModelFactory;
  }

  private FileModelProperties getFileModelProperties() {
    if (fileModelProperties == null) {
      fileModelProperties = GWT.create(FileModelProperties.class);
    }
    return fileModelProperties;
  }

  private SettingsAutoBeanFactory getSettingsAutoBeanFactory() {
    if (settingsAutoBeanFactory == null) {
      settingsAutoBeanFactory = GWT.create(SettingsAutoBeanFactory.class);
    }
    return settingsAutoBeanFactory;
  }

  private Storage getStorage() {
    if (storage == null) {
      storage = Storage.getLocalStorageIfSupported();
      Storage.addStorageEventHandler(new Handler() {
        @Override
        public void onStorageChange(StorageEvent event) {
          String key = event.getKey();
          String oldValue = event.getOldValue();
          String newValue = event.getNewValue();
          System.out.print(DateTimeFormat.getFormat(PredefinedFormat.TIME_MEDIUM).format(new Date()) + ": ");
          System.out.println("DesktopAppPresenter.onStorageChange: key=" + key + "\noldValue=" + oldValue
              + "\nnewValue=" + newValue);
        }
      });
    }
    return storage;
  }

  private String getUserKey(String userName, String key) {
    return userName + "." + key;
  }

  private void initializeSettings(Settings settings) {
    settings.setDesktopLayoutType(DesktopLayoutType.TILE);
  }

  private boolean isLocalStorageSupportChecked() {
    return isLocalStorageSupported != null;
  }

  private boolean isMoveEvent(Event<?> event) {
    return event instanceof HasMoveSupport ? ((HasMoveSupport) event).isMove() : false;
  }

}
