/**
 * Ext GWT 3.0.0-beta3 - Ext for GWT
 * Copyright(c) 2007-2011, Sencha, Inc.
 * licensing@sencha.com
 *
 * http://sencha.com/license
 */
package com.sencha.gxt.desktop.client.desktopapp.wordprocessor;

import com.google.web.bindery.event.shared.HandlerRegistration;
import com.sencha.gxt.core.client.util.Util;
import com.sencha.gxt.desktop.client.desktopapp.DesktopBus;
import com.sencha.gxt.desktop.client.desktopapp.event.AddFileModelEvent;
import com.sencha.gxt.desktop.client.desktopapp.event.AddFileModelHandler;
import com.sencha.gxt.desktop.client.desktopapp.event.RemoveFileModelEvent;
import com.sencha.gxt.desktop.client.desktopapp.event.RemoveFileModelHandler;
import com.sencha.gxt.desktop.client.desktopapp.event.UpdateFileModelEvent;
import com.sencha.gxt.desktop.client.desktopapp.event.UpdateFileModelHandler;
import com.sencha.gxt.desktop.client.desktopapp.persistence.FileModel;
import com.sencha.gxt.desktop.client.desktopapp.persistence.FileSystem;
import com.sencha.gxt.desktop.client.desktopapp.utility.Prompt;

public class WordProcessorPresenter implements Presenter {

  private DesktopBus desktopBus;
  private FileSystem fileSystem;
  private FileModel fileModel;

  private View view;
  private HandlerRegistration addFileModelRegistration;
  private AddFileModelHandler addFileModelHandler;
  private HandlerRegistration updateFileModelRegistration;
  private UpdateFileModelHandler updateFileModelHandler;
  private HandlerRegistration removeFileModelRegistration;
  private RemoveFileModelHandler removeFileModelHandler;
  private boolean isLocalUpdate;
  private String savedContent;

  public WordProcessorPresenter(DesktopBus desktopBus, FileSystem fileSystem, FileModel fileModel, View view) {
    this.desktopBus = desktopBus;
    this.fileSystem = fileSystem;
    this.fileModel = fileModel;
    this.view = view;
  }

  @Override
  public void bind() {
    getAddFileModelRegistration();
    getUpdateFileModelRegistration();
    getRemoveFileModelRegistration();
    setValue(getFileModel().getContent());
  }

  @Override
  public String getTitle() {
    return "Word Processor - " + getFileSystem().getPath(fileModel);
  }

  @Override
  public void onClose() {
    if (isSavedContent()) {
      getView().close();
    } else {
      Prompt.get().confirm("Close", "The document has unsaved changes. Would you like to discard them?",
          new Runnable() {
            @Override
            public void run() {
              getView().close();
            }
          });
    }
  }

  @Override
  public void onSave() {
    getFileModel().setContent(getView().getValue());
    updateFileModel();
  }

  @Override
  public void unbind() {
    getAddFileModelRegistration().removeHandler();
    getUpdateFileModelRegistration().removeHandler();
    getRemoveFileModelRegistration().removeHandler();
  }

  private AddFileModelHandler getAddFileModelHandler() {
    if (addFileModelHandler == null) {
      addFileModelHandler = new AddFileModelHandler() {
        @Override
        public void onAddFileModel(AddFileModelEvent addFileModelEvent) {
          onAddFileModelEvent(addFileModelEvent);
        }
      };
    }
    return addFileModelHandler;
  }

  private HandlerRegistration getAddFileModelRegistration() {
    if (addFileModelRegistration == null) {
      addFileModelRegistration = getDesktopBus().addAddFileModelHandler(getAddFileModelHandler());
    }
    return addFileModelRegistration;
  }

  private DesktopBus getDesktopBus() {
    return desktopBus;
  }

  private FileModel getFileModel() {
    return fileModel;
  }

  private FileSystem getFileSystem() {
    return fileSystem;
  }

  private RemoveFileModelHandler getRemoveFileModelHandler() {
    if (removeFileModelHandler == null) {
      removeFileModelHandler = new RemoveFileModelHandler() {
        @Override
        public void onRemoveFileModel(RemoveFileModelEvent removeFileModelEvent) {
          onRemoveFileModelEvent(removeFileModelEvent);
        }
      };
    }
    return removeFileModelHandler;
  }

  private HandlerRegistration getRemoveFileModelRegistration() {
    if (removeFileModelRegistration == null) {
      removeFileModelRegistration = getDesktopBus().addRemoveFileModelHandler(getRemoveFileModelHandler());
    }
    return removeFileModelRegistration;
  }

  private String getSavedContent() {
    return savedContent;
  }

  private UpdateFileModelHandler getUpdateFileModelHandler() {
    if (updateFileModelHandler == null) {
      updateFileModelHandler = new UpdateFileModelHandler() {
        @Override
        public void onUpdateFileModel(UpdateFileModelEvent updateFileModelEvent) {
          onUpdateFileModelEvent(updateFileModelEvent);
        }
      };
    }
    return updateFileModelHandler;
  }

  private HandlerRegistration getUpdateFileModelRegistration() {
    if (updateFileModelRegistration == null) {
      updateFileModelRegistration = getDesktopBus().addUpdateFileModelHandler(getUpdateFileModelHandler());
    }
    return updateFileModelRegistration;
  }

  private View getView() {
    return view;
  }

  private boolean isSavedContent() {
    return Util.equalWithNull(getSavedContent(), getView().getValue());
  }

  private void onAddFileModelEvent(AddFileModelEvent addFileModelEvent) {
    if (addFileModelEvent.isMove()) {
      // update title on any add to pick up parent name change
      updateTitle();
    }
  }

  private void onRemoveFileModelEvent(RemoveFileModelEvent removeFileModelEvent) {
    if (!removeFileModelEvent.isMove()) {
      FileModel fileModel = removeFileModelEvent.getFileModel();
      if (fileModel == getFileModel()) {
        getView().close();
      }
    }
  }

  private void onUpdateFileModelEvent(UpdateFileModelEvent updateFileModelEvent) {
    if (!isLocalUpdate) {
      FileModel fileModel = updateFileModelEvent.getFileModel();
      if (fileModel == getFileModel()) {
        updateContent(fileModel);
      }
      // update title on any update to pick up parent name change
      updateTitle();
    }
  }

  private void setSavedContent(String savedContent) {
    this.savedContent = savedContent;
  }

  private void setValue(String content) {
    getView().setValue(content);
    setSavedContent(content);
  }

  private void updateContent(FileModel fileModel) {
    final String content = fileModel.getContent();
    if (isSavedContent()) {
      setValue(content);
    } else {
      Prompt.get().confirm("Update",
          "The document has been updated elsewhere. Would you like to discard your changes?", new Runnable() {
            @Override
            public void run() {
              setValue(content);
            }
          });
    }
  }

  private void updateFileModel() {
    isLocalUpdate = true;
    getFileSystem().update(getFileModel());
    isLocalUpdate = false;
    setSavedContent(getFileModel().getContent());
  }

  private void updateTitle() {
    String newTitle = getTitle();
    getView().setTitle(newTitle);
  }

}
